@@ -1,6 +1,6 @@
 #!/usr/bin/env ocaml
 (* libguestfs
- * Copyright (C) 2016-2020 Red Hat Inc.
+ * Copyright (C) 2016-2022 Red Hat Inc.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -75,7 +75,9 @@ let () =
   ;;
 
 type os =
+  | Alma of int * int           (* major, minor *)
   | CentOS of int * int         (* major, minor *)
+  | CentOSStream of int         (* major *)
   | RHEL of int * int
   | Debian of int * string      (* version, dist name like "wheezy" *)
   | Ubuntu of string * string
@@ -138,6 +140,9 @@ let rec main () =
         | Aarch64 ->
            "/usr/share/edk2/aarch64/QEMU_EFI-pflash.raw",
            "/usr/share/edk2/aarch64/vars-template-pflash.raw"
+        | Armv7 ->
+           "/usr/share/edk2/arm/QEMU_EFI-pflash.raw",
+           "/usr/share/edk2/arm/vars-template-pflash.raw"
         | _ -> assert false in
 
       let vars_out = Sys.getcwd () // sprintf "%s.vars" tmpname in
@@ -345,13 +350,16 @@ Options:
 
 and os_of_string os ver =
   match os, ver with
+  | "alma", ver -> let maj, min = parse_major_minor ver in Alma (maj, min)
   | "centos", ver -> let maj, min = parse_major_minor ver in CentOS (maj, min)
+  | "centosstream", ver -> CentOSStream(int_of_string ver)
   | "rhel", ver -> let maj, min = parse_major_minor ver in RHEL (maj, min)
   | "debian", "6" -> Debian (6, "squeeze")
   | "debian", "7" -> Debian (7, "wheezy")
   | "debian", "8" -> Debian (8, "jessie")
   | "debian", "9" -> Debian (9, "stretch")
   | "debian", "10" -> Debian (10, "buster")
+  | "debian", "11" -> Debian (11, "bullseye")
   | "ubuntu", "10.04" -> Ubuntu (ver, "lucid")
   | "ubuntu", "12.04" -> Ubuntu (ver, "precise")
   | "ubuntu", "14.04" -> Ubuntu (ver, "trusty")
@@ -421,9 +429,15 @@ and filename_of_os os arch ext =
   | Fedora ver ->
      if arch = X86_64 then sprintf "fedora-%d%s" ver ext
      else sprintf "fedora-%d-%s%s" ver (string_of_arch arch) ext
+  | Alma (major, minor) ->
+     if arch = X86_64 then sprintf "alma-%d.%d%s" major minor ext
+     else sprintf "alma-%d.%d-%s%s" major minor (string_of_arch arch) ext
   | CentOS (major, minor) ->
      if arch = X86_64 then sprintf "centos-%d.%d%s" major minor ext
      else sprintf "centos-%d.%d-%s%s" major minor (string_of_arch arch) ext
+  | CentOSStream ver ->
+     if arch = X86_64 then sprintf "centosstream-%d%s" ver ext
+     else sprintf "centosstream-%d-%s%s" ver (string_of_arch arch) ext
   | RHEL (major, minor) ->
      if arch = X86_64 then sprintf "rhel-%d.%d%s" major minor ext
      else sprintf "rhel-%d.%d-%s%s" major minor (string_of_arch arch) ext
@@ -450,7 +464,9 @@ and string_of_os os arch = filename_of_os os arch ""
 (* This is what virt-builder called "os-version". *)
 and string_of_os_noarch = function
   | Fedora ver -> sprintf "fedora-%d" ver
+  | Alma (major, minor) -> sprintf "alma-%d.%d" major minor
   | CentOS (major, minor) -> sprintf "centos-%d.%d" major minor
+  | CentOSStream ver -> sprintf "centosstream-%d" ver
   | RHEL (major, minor) -> sprintf "rhel-%d.%d" major minor
   | Debian (ver, _) -> sprintf "debian-%d" ver
   | Ubuntu (ver, _) -> sprintf "ubuntu-%s" ver
@@ -460,25 +476,27 @@ and string_of_os_noarch = function
 
 (* Does virt-sysprep know how to sysprep this OS? *)
 and can_sysprep_os = function
-  | RHEL _ | CentOS _ | Fedora _ | Debian _ | Ubuntu _ -> true
+  | RHEL _ | Alma _ | CentOS _ | CentOSStream _ | Fedora _
+  | Debian _ | Ubuntu _ -> true
   | FreeBSD _ | Windows _ -> false
 
 and is_selinux_os = function
-  | RHEL _ | CentOS _ | Fedora _ -> true
+  | RHEL _ | Alma _ | CentOS _ | CentOSStream _ | Fedora _ -> true
   | Debian _ | Ubuntu _
   | FreeBSD _ | Windows _ -> false
 
 and needs_uefi os arch =
   match os, arch with
+  | Fedora _, Armv7
   | Fedora _, Aarch64
   | RHEL _, Aarch64 -> true
-  | RHEL _, _ | CentOS _, _ | Fedora _, _
+  | RHEL _, _ | Alma _, _ | CentOS _, _ | CentOSStream _, _ | Fedora _, _
   | Debian _, _ | Ubuntu _, _
   | FreeBSD _, _ | Windows _, _ -> false
 
 and get_virtual_size_gb os arch =
   match os with
-  | RHEL _ | CentOS _ | Fedora _
+  | RHEL _ | Alma _ | CentOS _ | CentOSStream _ | Fedora _
   | Debian _ | Ubuntu _
   | FreeBSD _ -> 6
   | Windows (10, _, _) -> 40    (* Windows 10 *)
@@ -489,7 +507,7 @@ and get_virtual_size_gb os arch =
 and make_kickstart os arch =
   match os with
   (* Kickstart. *)
-  | Fedora _ | CentOS _ | RHEL _ ->
+  | Fedora _ | Alma _ | CentOS _ | CentOSStream _ | RHEL _ ->
      let ks_filename = filename_of_os os arch ".ks" in
      Some (make_kickstart_common ks_filename os arch)
 
@@ -516,7 +534,8 @@ and make_kickstart_common ks_filename os arch =
   (* Fedora 34+ removes the "install" keyword. *)
   (match os with
    | Fedora n when n >= 34 -> ()
-   | RHEL (n, _) | CentOS (n, _) when n >= 9 -> ()
+   | RHEL (n, _)
+   | Alma (n, _) | CentOS (n, _) | CentOSStream n when n >= 9 -> ()
    | _ -> bpf "install\n";
   );
 
@@ -562,7 +581,7 @@ mouse generic
    *)
   (match os with
    | CentOS (8, _) ->
-      bpf "url --url=\"http://mirror.centos.org/centos-8/8/BaseOS/x86_64/os\"\n"
+      bpf "url --url=\"https://vault.centos.org/8.5.2111/BaseOS/x86_64/os/\"\n"
    | _ -> ()
   );
   bpf "\n";
@@ -578,7 +597,7 @@ part /boot --fstype=%s   --size=512         --asprimary
 part swap                --size=1024        --asprimary
 part /     --fstype=%s   --size=1024 --grow --asprimary
 " bootfs rootfs;
-   | CentOS _ | RHEL _ | Fedora _ ->
+   | Alma _ | CentOS _ | CentOSStream _ | RHEL _ | Fedora _ ->
       bpf "\
 zerombr
 clearpart --all --initlabel --disklabel=gpt
@@ -836,6 +855,11 @@ and make_unattend_iso os arch =
 
 and make_boot_media os arch =
   match os, arch with
+  | Alma (major, minor), X86_64 ->
+     (* UK mirror *)
+     Location (sprintf "http://mirror.cov.ukservers.com/almalinux/%d.%d/BaseOS/x86_64/kickstart/"
+                 major minor)
+
   | CentOS (major, _), Aarch64 ->
      (* XXX This always points to the latest CentOS, so
       * effectively the minor number is always ignored.
@@ -852,7 +876,11 @@ and make_boot_media os arch =
      Location "http://mirror.centos.org/centos-7/7/os/x86_64/"
 
   | CentOS (8, _), X86_64 ->
-     Location "http://mirror.centos.org/centos-8/8/BaseOS/x86_64/kickstart"
+     (* This is probably the last CentOS 8 release. *)
+     Location "https://vault.centos.org/8.5.2111/BaseOS/x86_64/kickstart/"
+
+  | CentOSStream ver, X86_64 ->
+     Location (sprintf "http://mirror.centos.org/centos/%d-stream/BaseOS/x86_64/os" ver)
 
   | Debian (_, dist), arch ->
      Location (sprintf "http://deb.debian.org/debian/dists/%s/main/installer-%s"
@@ -981,7 +1009,8 @@ The FreeBSD install is not automated.  Select all defaults, except:
 
 (* If the install is not automated and we need a graphical console. *)
 and needs_graphics = function
-  | CentOS _ | RHEL _ | Debian _ | Ubuntu _ | Fedora _ -> false
+  | Alma _ | CentOS _ | CentOSStream _ | RHEL _
+  | Debian _ | Ubuntu _ | Fedora _ -> false
   | FreeBSD _ | Windows _ -> true
 
 (* NB: Arguments do not need to be quoted, because we pass them
@@ -1034,6 +1063,10 @@ and make_virt_install_command os arch ks tmpname tmpout tmpefivars
       add "--machine=pseries";
       add "--cpu=power8";
       add "--vcpus=1"
+   | Armv7 ->
+      add "--arch=armv7l";
+      add "--machine=virt-2.11"; (* RHBZ#1633328, RHBZ#2003706 *)
+      add "--vcpus=1"
    | arch ->
       add (sprintf "--arch=%s" (string_of_arch arch));
       add "--vcpus=1"
@@ -1052,7 +1085,7 @@ and make_virt_install_command os arch ks tmpname tmpout tmpefivars
   (* --initrd-inject and --extra-args flags for Linux only. *)
   (match os with
    | Debian _ | Ubuntu _
-   | Fedora _ | RHEL _ | CentOS _ ->
+   | Fedora _ | RHEL _ | Alma _ | CentOS _ | CentOSStream _ ->
       let ks =
         match ks with None -> assert false | Some ks -> ks in
       add (sprintf "--initrd-inject=%s" ks);
@@ -1062,9 +1095,15 @@ and make_virt_install_command os arch ks tmpname tmpout tmpefivars
         | Debian _ | Ubuntu _ -> "auto"
         | Fedora n when n >= 34 ->
            sprintf "inst.ks=file:/%s" (Filename.basename ks)
-        | RHEL (n, _) | CentOS (n, _) when n >= 9 ->
+        | Alma (major, _) ->
+           (* This is only required because of missing osinfo-db data.
+            * https://bugs.almalinux.org/view.php?id=127
+            * Once this is fixed, do the same as CentOS below.
+            *)
+           sprintf "inst.ks=file:/%s inst.repo=http://repo.almalinux.org/almalinux/%d/BaseOS/x86_64/os/" (Filename.basename ks) major
+        | RHEL (n, _) | CentOS (n, _) | CentOSStream n when n >= 9 ->
            sprintf "inst.ks=file:/%s" (Filename.basename ks)
-        | Fedora _ | RHEL _ | CentOS _ ->
+        | Fedora _ | RHEL _ | CentOS _ | CentOSStream _ ->
            sprintf "ks=file:/%s" (Filename.basename ks)
         | FreeBSD _ | Windows _ -> assert false in
       let proxy =
@@ -1072,15 +1111,19 @@ and make_virt_install_command os arch ks tmpname tmpout tmpefivars
         match p with
         | None ->
            (match os with
-            | Fedora _ | RHEL _ | CentOS _ | Ubuntu _ -> ""
+            | Fedora _ | RHEL _ | Alma _ | CentOS _ | CentOSStream _
+            | Ubuntu _ -> ""
             | Debian _ -> "mirror/http/proxy="
             | FreeBSD _ | Windows _ -> assert false
            )
         | Some p ->
            match os with
            | Fedora n when n >= 34 -> sprintf "inst.proxy=" ^ p
-           | RHEL (n, _) | CentOS (n, _) when n >= 9 -> "inst.proxy=" ^ p
-           | Fedora _ | RHEL _ | CentOS _ -> "proxy=" ^ p
+           | RHEL (n, _)
+           | Alma (n, _) | CentOS (n, _) | CentOSStream n when n >= 9 ->
+              "inst.proxy=" ^ p
+           | Fedora _ | RHEL _ | Alma _ | CentOS _ | CentOSStream _ ->
+              "proxy=" ^ p
            | Debian _ | Ubuntu _ -> "mirror/http/proxy=" ^ p
            | FreeBSD _ | Windows _ -> assert false in
 
@@ -1137,7 +1180,9 @@ and os_variant_of_os ?(for_fedora = false) os arch =
   if not for_fedora then (
     match os with
     | Fedora ver -> sprintf "fedora%d" ver
+    | Alma (major, _) -> sprintf "almalinux%d" major
     | CentOS (major, minor) -> sprintf "centos%d.%d" major minor
+    | CentOSStream ver -> sprintf "centosstream%d" ver
     | RHEL (major, minor) -> sprintf "rhel%d.%d" major minor
     | Debian (ver, _) -> sprintf "debian%d" ver
     | Ubuntu (ver, _) -> sprintf "ubuntu%s" ver
@@ -1159,13 +1204,18 @@ and os_variant_of_os ?(for_fedora = false) os arch =
     | Fedora _, (PPC64|PPC64le) -> "fedora22"
     | Fedora ver, _ when ver <= 23 ->
        sprintf "fedora%d" ver
-    | Fedora _, _ -> "fedora26" (* max version known in Fedora 28 *)
+    | Fedora _, _ -> "fedora34" (* max version known in Fedora 34 *)
+
+    | Alma (major, _), _ -> sprintf "almalinux%d" major
 
     | CentOS (8, _), _ -> "rhel8.0" (* temporary until osinfo updated *)
     | CentOS (major, minor), _ when (major, minor) <= (7,0) ->
        sprintf "centos%d.%d" major minor
     | CentOS _, _ -> "centos7.0" (* max version known in Fedora 31 *)
 
+    | CentOSStream 8, _ -> "rhel8.0" (* temporary until osinfo updated *)
+    | CentOSStream _, _ -> "rhel8.0" (* min known version is 8 *)
+
     | RHEL (6, minor), _ when minor <= 8 ->
        sprintf "rhel6.%d" minor
     | RHEL (6, _), _ -> "rhel6.9" (* max version known in Fedora 29 *)
@@ -1205,8 +1255,8 @@ and kernel_cmdline_of_os os arch =
      "console=tty0 console=ttyAMA0,115200 rd_NO_PLYMOUTH"
   | (Debian _|Fedora _|Ubuntu _), (PPC64|PPC64le) ->
      "console=tty0 console=hvc0 rd_NO_PLYMOUTH"
-  | (RHEL _|CentOS _), PPC64
-  | (RHEL _|CentOS _), PPC64le ->
+  | (RHEL _ | Alma _ | CentOS _ | CentOSStream _), PPC64
+  | (RHEL _ | Alma _ | CentOS _ | CentOSStream _), PPC64le ->
      "console=tty0 console=ttyS0,115200 rd_NO_PLYMOUTH"
 
   | FreeBSD _, _ | Windows _, _ -> assert false
@@ -1231,7 +1281,8 @@ and make_postinstall os arch =
          g#write "/etc/yum.repos.d/download.devel.redhat.com.repo" yum_conf
      )
 
-  | RHEL _ | Fedora _ | CentOS _ | FreeBSD _ | Windows _ -> None
+  | RHEL _ | Fedora _ | Alma _ | CentOS _ | CentOSStream _
+  | FreeBSD _ | Windows _ -> None
 
 and make_rhel_yum_conf major minor arch =
   let buf = Buffer.create 4096 in
@@ -1379,10 +1430,18 @@ and make_index_fragment os arch index_fragment output nvram revision
 
 and long_name_of_os os arch =
   match os, arch with
+  | Alma (major, minor), X86_64 ->
+     sprintf "AlmaLinux %d.%d" major minor
+  | Alma (major, minor), arch ->
+     sprintf "AlmaLinux %d.%d (%s)" major minor (string_of_arch arch)
   | CentOS (major, minor), X86_64 ->
      sprintf "CentOS %d.%d" major minor
   | CentOS (major, minor), arch ->
      sprintf "CentOS %d.%d (%s)" major minor (string_of_arch arch)
+  | CentOSStream ver, X86_64 ->
+     sprintf "CentOS Stream %d" ver
+  | CentOSStream ver, arch ->
+     sprintf "CentOS Stream %d (%s)" ver (string_of_arch arch)
   | Debian (ver, dist), X86_64 ->
      sprintf "Debian %d (%s)" ver dist
   | Debian (ver, dist), arch ->
@@ -1425,8 +1484,12 @@ and notes_of_os os arch nvram =
   add "";
 
   (match os with
+   | Alma _ ->
+      add "This AlmaLinux image contains only unmodified @Core group packages."
    | CentOS _ ->
       add "This CentOS image contains only unmodified @Core group packages."
+   | CentOSStream _ ->
+      add "This CentOS Stream image contains only unmodified @Core group packages."
    | Debian _ ->
       add "This is a minimal Debian install."
    | Fedora _ ->
